/*BEGIN OCO COPYRIGHT
 *************************************************************************
 *
 * IBM Confidential
 * OCO Source Materials
 * 5724-L01, 5655-N53, 5724-I82, 5655-R15
 * (C) Copyright IBM Corporation 2006.
 * The source code for this program is not published or otherwise
 * divested of its trade secrets, irrespective of what has been
 * deposited with the U.S. Copyright Office.
 *
 **************************************************************************
 *END OCO COPYRIGHT
 */
 
package com.ibm.websphere.samples.databinding.util;


import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.sdo.EProperty;
import org.eclipse.emf.ecore.util.EcoreUtil;

import com.ibm.websphere.bo.BOFactory;
import com.ibm.websphere.samples.databinding.exceptions.DelimitedHandlerException;
import com.ibm.websphere.samples.databinding.log.HandlerUtil;
import com.ibm.websphere.sca.ServiceManager;
import commonj.sdo.DataObject;
import commonj.sdo.Property;
import commonj.sdo.Type;

/**
 * The DelimitedDataHandler converts a delimited stream of data to a 
 * data object and vice versa. 
 * A comma delimited stream for a customer business object which contain properties
 * id, firstName, lastName, contained object address with id looks as 
 * follows Customer,10,John,Doe,address,101.
 * There are configuration properties like Delimiter, TargetNamespace that can be
 * passed to the handler.
 * <p>
 * @ibm-api
 */
public class DelimitedDataHandler
{
    /** Automatically generated javadoc for: BUFFERSIZE */
    private static final int BUFFERSIZE = 256;
    //	BEGIN COPYRIGHT VARIABLE
	public static final String COPYRIGHT = "(C) Copyright IBM Corporation 2006.";	//$NON-NLS-$1
	// END COPYRIGHT VARIABLE
	 
    //Configuration properties
    //These parameters can be set by the client to indicate what the stream would be like
    private char charDelimiter = ',';
    private String delimiter = ",";  //$NON-NLS-1$
    private boolean isCharDelimiter = true;
    private char charEscape = '\\';
    private String escape = "\\";  //$NON-NLS-1$
    private String targetNamespace = "http://www.ibm.com/crm";  //$NON-NLS-1$
    
    private char[] charBuffer = null;
    private StringBuffer DOStringBuffer = new StringBuffer();
    private final String emptyValue = "";  //$NON-NLS-1$
    private int currentIndex = 0;

    private boolean endOfDOString = false;
    private static int startIndex = 0;
    private static final int attrValueBufferSize = 256;
    private static final char TAB_CHAR = '\t';    

    private static ServiceManager serviceManager = new ServiceManager();
    private static final BOFactory boFactory = (BOFactory) serviceManager.locateService("com/ibm/websphere/bo/BOFactory");  //$NON-NLS-1$

    private static final String CLASSNAME = DelimitedDataHandler.class.getName();
    private static final Logger LOGGER = HandlerUtil.initLogger(CLASSNAME);


    /**
     * Converts a delimited byte stream to a data object. It uses a hashmap of 
     * configuration properties. The properties are Delimiter,
     * Escape and TargetNamespace. The default values for these properties
     * are set to  ",", "//" and "http://www.ibm.com/crm" respectively.
     * If the config is null, the default values are used for parsing the
     * stream. The default values can be overridden by passing them in the
     * config hashmap.
     * 
     * @param bytes input bytes to be converted
     * @param config map of configuration properties
     * @return dataobject that corresponds to the byte stream
     * @throws DelimitedHandlerException The bytes can be read
     */
    public DataObject getDataObject(byte[] bytes,
                                    HashMap config)
    throws DelimitedHandlerException
        
    {
        logEntry("getDataObject(byte[], HashMap)");  //$NON-NLS-1$
        HandlerUtil.logConfig(LOGGER, CLASSNAME, "getDataObject(byte[], HashMap)", config);  //$NON-NLS-1$

        setupConfig(config);
        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
        try {
            outputStream.write(bytes);
        } catch (IOException e) {
            throw new DelimitedHandlerException("Error when reading the serialized data", e);  //$NON-NLS-1$
        }
        String dataObjectString = outputStream.toString();

        LOGGER.logp(Level.FINEST, CLASSNAME, "getDataObject(byte[], HashMap)", //$NON-NLS-1$
                    "Incoming byte array is : \"{0}\"", dataObjectString);  //$NON-NLS-1$

        charBuffer = new char[dataObjectString.length()];
        dataObjectString.getChars(0, charBuffer.length, charBuffer, 0);
        String dataObjectName = getNextToken();
        DataObject dataObject = boFactory.create(targetNamespace, dataObjectName);

        LOGGER.logp(Level.FINER, CLASSNAME, "getDataObject(byte[], HashMap)", //$NON-NLS-1$
                    "Name of the created DataObject: \"{0}\"", targetNamespace+"//"+dataObjectName);  //$NON-NLS-1$
        
        LOGGER.logp(Level.FINER, CLASSNAME, "getDataObject(byte[], HashMap)", //$NON-NLS-1$
                "Created the DataObject: \"{0}\"", dataObject);  //$NON-NLS-1$
        // parse through the main DataObject attributes
        parseAllProperties(dataObject);
        HandlerUtil.logDataObject(LOGGER, CLASSNAME, "getDataObject(byte[], HashMap)", //$NON-NLS-1$
                                  dataObject);

        logExit("getDataObject(byte[], HashMap)");  //$NON-NLS-1$
        return dataObject;
    }


    /**
     * Converts a dataobject to a delimited byte stream. It uses a hashmap of
     * configuration properties. The properties are Delimiter,
     * Escape and TargetNamespace. The default values for these properties
     * are set to  ",", "//" and "http://www.ibm.com/crm" respectively.
     * If the config is null, the default values are used for serializing the
     * dataobject to byte stream. The default values can be overridden by passing them in the
     * config hashmap.
     *
     * @param config Map of configuration properties
     * @return Delimited byte stream that corresponds to the dataobject
     * @throws DelimitedHandlerException The bytes cannot be read
     * @param dataObject 
     */
    public byte[] getBytesFromDataObject(DataObject dataObject,
                                         HashMap config)
        throws DelimitedHandlerException
    {
        logEntry("getBytesFromDataObject(DataObject, HashMap)");  //$NON-NLS-1$
        HandlerUtil.logDataObject(LOGGER, CLASSNAME, "getBytesFromDataObject(DataObject, HashMap)", //$NON-NLS-1$
                                  dataObject);
        HandlerUtil.logConfig(LOGGER, CLASSNAME, "getBytesFromDataObject(DataObject, HashMap)", config);  //$NON-NLS-1$

        setupConfig(config);
        Type type = dataObject.getType();
        String dataObjName = type.getName();
        setSimpleToken(true, dataObjName);
        serializeAllProperties(dataObject);
        String dataObjectString = DOStringBuffer.toString();

        LOGGER.logp(Level.FINER, CLASSNAME, "getStringFromDataObject(DataObject dataObject, HashMap map)", //$NON-NLS-1$
                    "Created the Stream, the Stream is : \"{0}\"", dataObjectString);  //$NON-NLS-1$

        logExit("getBytesFromDataObject(DataObject, HashMap)");  //$NON-NLS-1$
        return dataObjectString.getBytes();
    }


    /**
     * Converts a delimited string to a data object. It uses a hashmap of
     * configuration properties. The properties are Delimiter,
     * Escape and TargetNamespace. The default values for these properties
     * are set to  ",", "//" and "http://www.ibm.com/crm" respectively.
     * If the config is null, the default values are used for parsing the
     * stream. The default values can be overridden by passing them in the
     * config hashmap.
     *
     * @param config Map of configuration properties
     * @return Dataobject that corresponds to the byte stream
     * @throws DelimitedHandlerException The string can be read
     * @param serializedData 
     */
    public DataObject getDataObject(String serializedData,
                                    HashMap config)
        throws DelimitedHandlerException
    {
        logEntry("getDataObject(String, HashMap)");  //$NON-NLS-1$
        LOGGER.logp(Level.FINER, CLASSNAME, "getDataObject(String, HashMap)", //$NON-NLS-1$
                    "Incoming string is : \"{0}\"", serializedData);  //$NON-NLS-1$
        HandlerUtil.logConfig(LOGGER, CLASSNAME, "getDataObject(String, HashMap)", config);  //$NON-NLS-1$

        setupConfig(config);
        charBuffer = new char[serializedData.length()];
        serializedData.getChars(0, charBuffer.length, charBuffer, 0);

        // Get DataObject name - assumes it is the first token
        //  in the delmited string
        String dataObjectName = getNextToken();
        DataObject dataObject = boFactory.create(targetNamespace, dataObjectName);

        LOGGER.logp(Level.FINER, CLASSNAME, "getDataObjectFromString(String, HashMap)", //$NON-NLS-1$
                    "Created a DataObject: \"{0}\"", dataObjectName);  //$NON-NLS-1$

        // parse through the main DataObject attributes
        parseAllProperties(dataObject);

        HandlerUtil.logDataObject(LOGGER, CLASSNAME, "getDataObject(String, HashMap)", dataObject);  //$NON-NLS-1$
        logExit("getDataObject(String, HashMap)");  //$NON-NLS-1$
        return dataObject;
    }


    /**
     * Converts a dataobject to a delimited string. It uses a hashmap of
     * configuration properties. The properties are Delimiter,
     * Escape and TargetNamespace. The default values for these properties
     * are set to  ",", "//" and "http://www.ibm.com/crm" respectively.
     * If the config is null, the default values are used for serializing the
     * dataobject to string. The default values can be overridden by passing them in the
     * config hashmap.
     *
     * @param config Map of configuration properties
     * @return Delimited string that corresponds to the dataobject
     * @throws DelimitedHandlerException
     * @param dataObject 
     */
    public String getStringFromDataObject(DataObject dataObject,
                                          HashMap config)
        throws DelimitedHandlerException
    {
        logEntry("getStringFromDataObject(DataObject, HashMap)");  //$NON-NLS-1$
        HandlerUtil.logDataObject(LOGGER, CLASSNAME, "getStringFromDataObject(DataObject, HashMap)", dataObject);  //$NON-NLS-1$
        HandlerUtil.logConfig(LOGGER, CLASSNAME, "getStringFromDataObject(DataObject, HashMap)", config);  //$NON-NLS-1$

        setupConfig(config);
        Type type = dataObject.getType();
        String dataObjName = type.getName();
        setSimpleToken(true, dataObjName);
        serializeAllProperties(dataObject);
        String dataObjectStr = DOStringBuffer.toString();

        LOGGER.logp(Level.FINER, CLASSNAME, "getStringFromDataObject(DataObject dataObject, HashMap map)", //$NON-NLS-1$
                    "Created the Stream, the Stream is : \"{0}\"", dataObjectStr);  //$NON-NLS-1$

        logExit("getStringFromDataObject(DataObject dataObject, HashMap config)");  //$NON-NLS-1$
        return dataObjectStr;
    }


    /**
     * Loops through all the properties in a dataObject and adds the property
     * values to the DOStringBuffer separated by the delimiter.
     *
     * @param dataObject DataObject which contains the data to be added to the
     * delimited string.
     * @throws com.ibm.websphere.samples.databinding.exceptions.DelimitedHandlerException 
     */
    protected void serializeAllProperties(DataObject dataObject)
        throws DelimitedHandlerException
    {
        logEntry("serializeAllProperties(DataObject)");  //$NON-NLS-1$

        if(dataObject == null)
        {
            logExit("serializeAllProperties(DataObject)");  //$NON-NLS-1$
            return;
        }

        Type type = dataObject.getType();                
        // Loop through the properties in the dataObject to add
        // the value of each to the delimited string.
        Iterator propertyIterator = type.getProperties().iterator();
        while (propertyIterator.hasNext()) {
            Property property = (Property) propertyIterator.next();
            boolean isContainment = property.isContainment();
            boolean isMany = property.isMany();
            String propertyName = property.getName();           

            // Child dataobject is multiple cardinality, get the dataObject
            // and for each dataObject call serializeAllProperties method recursively       
            if (isContainment && isMany) {
                LOGGER.logp(Level.FINER, CLASSNAME, "serializeAllProperties(DataObject)", //$NON-NLS-1$
                            "Serializing complex type list property : \"{0}\"", propertyName);  //$NON-NLS-1$
                serializeChildDataObjectList(dataObject, property, propertyName);
                continue;
            }
            // property is single cardinality child object, get the dataObject
            // and call serializeAllProperties method on that dataobject
            if (isContainment && !isMany) {
                LOGGER.logp(Level.FINER, CLASSNAME, "serializeAllProperties(DataObject)", //$NON-NLS-1$
                            "Serializing complex type property : \"{0}\"", propertyName);  //$NON-NLS-1$
                serializeChildDataObject(dataObject, propertyName);
                continue;

            }
            //property is a simple type list
            if (!isContainment && isMany) {
                serializeSimpleTypeList(dataObject, property, propertyName);
                continue;
            }
            //property is simple type
            if (!isContainment && !isMany) {
                LOGGER.logp(Level.FINER, CLASSNAME, "serializeAllProperties(DataObject)", //$NON-NLS-1$
                            "Serializing simple type property : \"{0}\"", propertyName);  //$NON-NLS-1$
                if (dataObject.isSet(property.getName())) {
                    String strValue = dataObject.getString(propertyName);
                    setSimpleToken(true, strValue);
                    continue;
                }
                // no value was set so set it to ""
                setSimpleToken(true, emptyValue);
                continue;
            }

        } // end iterator
        logExit("serializeAllProperties(DataObject)");  //$NON-NLS-1$
    } // end setAttrList

    /**
     * First serialize the size of the list, next serialize the name of the property
     * corresponding to the child data object from the list and serialize its
     * property values to a delimited string.
     *
     * @param dataObject The property value to be serialized is part of this dataobject
     * @param property The value of this property is serialized
     * @param propertyName Property name corresponding to the data object.
     * @throws com.ibm.websphere.samples.databinding.exceptions.DelimitedHandlerException 
     */
    protected void serializeChildDataObjectList(DataObject dataObject,
                                           Property property,
                                           String propertyName)
        throws DelimitedHandlerException
    {
        int objectCount = dataObject.getList(property).size();
    
        setSimpleToken(false, Integer.toString(objectCount));

        Iterator childDOIterator = dataObject.getList(property).iterator();
        while (childDOIterator.hasNext()) {
            DataObject childDataObject = (DataObject) childDOIterator.next();
            if(childDataObject == null) {
                return;
            }
            setSimpleToken(true, propertyName);
            serializeAllProperties(childDataObject);
        }
    }

    
    /**
     * Serialize the property values of child data object to a delimited string
     * 
     * @param dataObject The property value to be serialized is part of this dataobject
     * @param propertyName The value of this property is serialized
     */
    protected void serializeChildDataObject(DataObject dataObject,
                                            String propertyName)
    	throws DelimitedHandlerException
    {        
        DataObject singleCardDataObject = dataObject.getDataObject(propertyName);
        if(singleCardDataObject == null) {
            return;
        }
        setSimpleToken(true, propertyName);
        serializeAllProperties(singleCardDataObject);
    }
    

    /**
     * Serialize out the size of the list and then the values in the list.
     *
     * @param dataObject The DataObject that contains the list of simple type
     * @param property The property that corresponds to the list of simple type
     * @param propertyName The name of the above property
     * @throws com.ibm.websphere.samples.databinding.exceptions.DelimitedHandlerException 
     */
    protected void serializeSimpleTypeList(DataObject dataObject,
                                           Property property,
                                           String propertyName)
        throws DelimitedHandlerException
    {
        logEntry("serializeSimpleTypeList(DataObject, Property, String)");  //$NON-NLS-1$
        LOGGER.logp(Level.FINER, CLASSNAME, "serializeSimpleTypeList(DataObject, Property, String)", //$NON-NLS-1$
                    "Simple Type list property name is : \"{0}\"", propertyName);  //$NON-NLS-1$

        List list = dataObject.getList(propertyName);
        //list is empty so just create an empty token (,,) in the stream
        if (list.size() == 0) {
            setSimpleToken(true, emptyValue);
            logExit("serializeSimpleTypeList(DataObject, Property, String)");  //$NON-NLS-1$
            return;
        }
        
        //serialize the list size
        setSimpleToken(false, Integer.toString(list.size()));
        //serialize the values
        try {
            EDataType dataType = (EDataType) ((EProperty) property).getEStructuralFeature().getEType();
            int size = list.size();
            for (int j = 0; j < size; j++) {
                Object value = list.get(j);
                String strValue = null;
                //the value can be string or some other type so convert the
                //value to a string so it can be tokenized
                if (dataType != null) {
                    strValue = EcoreUtil.convertToString(dataType, value);
                }
                setSimpleToken(true, strValue);
            }
        } catch (Exception e) {
            String message = "An error was encountered while setting list of attributes in the DataObject "  //$NON-NLS-1$
                             + dataObject.getType().getName() + " with the attribute Name " + propertyName;  //$NON-NLS-1$
            LOGGER.logp(Level.SEVERE, CLASSNAME, "serializeSimpleTypeList(DataObject, Property, String)", message);  //$NON-NLS-1$
            throw new DelimitedHandlerException(message, e.getCause());
        }
        logExit("serializeSimpleTypeList(DataObject, Property, String)");  //$NON-NLS-1$
    }


    /**
     * Parses the stream to construct the dataObject. This method is called
     * recursively.
     *
     * @param dataObject DataObject that is created from the incoming stream of data.
     * @throws com.ibm.websphere.samples.databinding.exceptions.DelimitedHandlerException 
     */
    protected void parseAllProperties(DataObject dataObject)
        throws DelimitedHandlerException
    {
        logEntry("parseAllProperties(DataObject)");  //$NON-NLS-1$
        // Assumption is made that if this method is called, there should
        // be some part of the string left to parse, otherwise we generate
        // an error.
        if (endOfDOString || dataObject == null) {
            String message = "The Input Stream has no tokens in it or the DataObject is null while invoking method "  //$NON-NLS-1$
                             + "parseAllProperties";  //$NON-NLS-1$
            LOGGER.logp(Level.SEVERE, CLASSNAME, "parseAttributeList(DataObject dataObject)", message);  //$NON-NLS-1$
            throw new DelimitedHandlerException(message, null);

        }
        Type type = dataObject.getType();
        Iterator propertyIterator = type.getProperties().iterator();
        while (propertyIterator.hasNext()) {
            Property property = (Property) propertyIterator.next();
            String propertyName = property.getName();            
            String propertyValue = getNextToken();
            boolean isContainment = property.isContainment();
            boolean isMany = property.isMany();
            String propertyTypeName = property.getType().getName();
            //the property is a list of child objects
            if (isContainment && isMany) {
                parseChildDataObjectList(dataObject, propertyName, propertyValue);
                continue;
            }
            //the property is a single child object
            if (isContainment && !isMany) {

                parseChildDataObject(dataObject, propertyName, propertyValue, propertyTypeName);
                continue;
            }
            //the property is a list of simple types
            if (!isContainment && isMany) {
                parseSimpleTypeList(dataObject, propertyValue, propertyName);
                continue;
            }
            // For simple (String) attribute, we check to see if the
            // token is null. Otherwise, set the attribute value with the token
            if (!isContainment && !isMany) {
                if (propertyValue == null || propertyValue.equals(emptyValue)) {
                    dataObject.unset(propertyName);
                    continue;
                }
                dataObject.setString(propertyName, propertyValue);
            }
        }
        logExit("parseAllProperties(DataObject)");  //$NON-NLS-1$
    } // end parseAttributeList


    /**
     * Populates a list of child data obejcts based on token values in the delimited
     * string. Expects all objects in the child container to be of the same
     * type.
     *
     * @param parentDataObject Parent DataObject.
     * @param propertyName Name of the property for the child data object
     * @param objectCountString Number of instances of the child object as parsed from the
     *            delimited string.
     * @throws com.ibm.websphere.samples.databinding.exceptions.DelimitedHandlerException 
     */
    protected void parseChildDataObjectList(DataObject parentDataObject,
                                            String propertyName,
                                            String objectCountString)
        throws DelimitedHandlerException

    {
        logEntry("parseChildDataObjectList(DataObject, String, String)");  //$NON-NLS-1$
        LOGGER.logp(Level.FINE, CLASSNAME, "parseChildDataObjectList(DataObject, String, String)",  //$NON-NLS-1$
                    "Parsing complex type list \"{0}\"", propertyName);  //$NON-NLS-1$
        
        
        try {
            //if the count read is empty or 0, then just return
            if (" ".equals(objectCountString)) {  //$NON-NLS-1$
                LOGGER.logp(Level.FINE, CLASSNAME, "parseChildDataObjectList(DataObject,String, String)",  //$NON-NLS-1$
                            "The count for multiple cardinality datObject \"{0}\" is empty string", propertyName);  //$NON-NLS-1$
                return;
            }           
            int objectCount = Integer.parseInt(objectCountString);
            if (objectCount == 0) {
                LOGGER.logp(Level.FINE, CLASSNAME, "parseChildDataObjectList(DataObject,String, String)",  //$NON-NLS-1$
                            "The count for multiple cardinality datObject \"{0}\" is 0", propertyName);  //$NON-NLS-1$
                return;
            }

            if (objectCount > 0) {
                // There is at least one instance of the object in the
                //  delimited string
                DataObject childDataObject = null;

                // For each instance of the child object, parse the
                //  property list, and add the object to the list
                //  in the parent.
                for (int objectIndex = 0; objectIndex < objectCount; objectIndex++) {
                    // The first call to getNextToken will remove the BOName.
                    // we need to get the next token , so that we can advance in
                    // the stream. but we don't want to do anything with this token.
                    getNextToken();
                    childDataObject = parentDataObject.createDataObject(propertyName);                    
                    parseAllProperties(childDataObject);

                    LOGGER.logp(Level.FINER, CLASSNAME, "parseChildDataObjectList(DataObject, String, String)",  //$NON-NLS-1$
                                "Created ChildDataObject: \"{0}\"", propertyName);  //$NON-NLS-1$
                }
            }
        } catch (Exception e) {
            String message = "An exception was encountered while trying to access or create attribute " + propertyName  //$NON-NLS-1$
                             + " on DataObject " + parentDataObject.getClass().getName();  //$NON-NLS-1$
            LOGGER.logp(Level.SEVERE, CLASSNAME,
                        "getMultipleCard(DataObject parentDataObject,String propertyName,String objectCountString)",  //$NON-NLS-1$
                        message);
            throw new DelimitedHandlerException(message, e.getCause());
        }
        logExit("parseChildDataObjectList(DataObject, String, String)");  //$NON-NLS-1$
    } // end getMultipleCard


    /**
     * Populates a list of Simple types in the data object.
     * @param dataObject DataObject that is created from the delimited stream
     * @param listCount  size of the simple type list as read from the stream
     * @param propertyName name of the property for this simple type list
     * @throws DelimitedHandlerException
     */
    protected void parseSimpleTypeList(DataObject dataObject,
                                       String listCount,
                                       String propertyName)
        throws DelimitedHandlerException
    {
        logEntry("parseSimpleTypeList(DataObject, String, String)");  //$NON-NLS-1$
        LOGGER.logp(Level.FINE, CLASSNAME, "parseSimpleTypeList(DataObject, String, String)",  //$NON-NLS-1$
                    "Parsing simple type list \"{0}\"", propertyName);  //$NON-NLS-1$

        if (listCount == null || listCount.equals(emptyValue)) {
            LOGGER.logp(Level.FINE, CLASSNAME, "parseSimpleTypeList(DataObject, String, String)",  //$NON-NLS-1$
                        "The simple type list \"{0}\" is empty", propertyName);  //$NON-NLS-1$

            return;
        }
        int objectCount = Integer.parseInt(listCount);
        if (objectCount == 0) {
            LOGGER.logp(Level.FINE, CLASSNAME, "parseSimpleTypeList(DataObject, String, String)",  //$NON-NLS-1$
                        "The simple type list \"{0}\" has 0 elements", propertyName);  //$NON-NLS-1$
            return;
        }
        if (objectCount < 0) {
            String message = "The simple type list " + propertyName + " must have more than 0 elements";  //$NON-NLS-2$  //$NON-NLS-1$
            LOGGER.logp(Level.SEVERE, CLASSNAME, "parseSimpleTypeList(DataObject, String, String)", message);  //$NON-NLS-1$
            throw new DelimitedHandlerException(message, null);

        }

        try {
            List outputList = dataObject.getList(propertyName);
            EStructuralFeature eFeature = ((EObject) dataObject).eClass().getEStructuralFeature(propertyName);
            EDataType outEType = (EDataType) eFeature.getEType();

            for (int j = 0; j < objectCount; j++) {
                // first get the token from the stream .
                String strValue = getNextToken();
                // Convert the value to the list type and add to list.
                Object convertedValue = EcoreUtil.createFromString(outEType, strValue);
                outputList.add(j, convertedValue);
            }
        } catch (Exception ex) {
            String message = "An exception was encountered while trying to access or create attribute " + propertyName  //$NON-NLS-1$
                             + "on DataObject " + dataObject.getClass().getName();  //$NON-NLS-1$
            LOGGER.logp(Level.SEVERE, CLASSNAME,
                        "setListOnDataObject(DataObject dataObject,String listCount,String propertyName)", message);  //$NON-NLS-1$
            throw new DelimitedHandlerException(message, ex.getCause());
        }
        logExit("parseSimpleTypeList(DataObject, String, String)");  //$NON-NLS-1$
    }


    /**
     * Populates a single cardinality child business object with the token
     * values found in the delimited string.
     *
     * @throws DelimitedHandlerException
     * @param parent
     * @param propertyName
     * @param propertyTypeName
     * @param propertyValue 
     */
    protected void parseChildDataObject(DataObject parent,
                                        String propertyName,
                                        String propertyValue,
                                        String propertyTypeName)
        throws DelimitedHandlerException
    {
        logEntry("parseChildDataObject(DataObject, String, String, String)");  //$NON-NLS-1$
        LOGGER.logp(Level.FINE, CLASSNAME, "parseChildDataObject(DataObject, String, String, String)",  //$NON-NLS-1$
                    "Parsing child data object \"{0}\"", propertyName);  //$NON-NLS-1$
        
        //propertyValue is either childBOName or is ""
        if ("".equals(propertyValue)) {  //$NON-NLS-1$
            return;
        }

        String childBOName = propertyValue;
        try {
            // Check the object count token
            // If it does not match "1", assume that the child object should be null
            if (!propertyName.equalsIgnoreCase(childBOName)) {
                return;
            }
            // need to create the ChildBO on the ParentBO
            DataObject childBO = parent.createDataObject(propertyName);
            parseAllProperties(childBO);
        } catch (Exception e) {            
            String message = "An exception was encountered while trying to access or create Child DataObject"  //$NON-NLS-1$
                             + propertyName + " on DataObject " + parent.getClass().getName();  //$NON-NLS-1$
            LOGGER.logp(Level.SEVERE, CLASSNAME, "parseChildDataObject(DataObject, String, String, String)", message);  //$NON-NLS-1$
            throw new DelimitedHandlerException(message, e.getCause());
        }
        logExit("parseChildDataObject(DataObject, String, String, String)");  //$NON-NLS-1$
    } 


    /**
     * Parses the next token from BOString based on the next appearance of the
     * configured delimiter. Also deletes the escape character from the token
     * value where it immediately preceeds the delimiter.
     * 
     * @return String value that is the next token that appers in the mBOString.
     */
    protected String getNextToken()
    {
        logEntry("getNextToken()");  //$NON-NLS-1$
        if (currentIndex >= charBuffer.length) {
            //throw new CW_EndOfBOException(" Already reached end of BO while
            // expecting more token ");
        }

        String result = null;

        // Check to see if the Delimiter can be
        // represented as a single character. If so, follow
        // a different code path for efficiency purposes.
        // charDelimiterRead = efficient because parsing char
        // stringDelmiterRead = less efficient because parsing string
        if (isCharDelimiter) {
            result = charDelimiterRead();
        }
        else {
            result = stringDelimiterRead();
        }
        logExit("getNextToken()");  //$NON-NLS-1$
        return result;
    } // end getNextToken


    /**
     * Override the default configuration properties on the Handler based on the configuration
     * that is passed in.
     *
     * @param config 
     */

    protected void setupConfig(HashMap config)
    {
        logEntry("setupConfig(HashMap)");  //$NON-NLS-1$

        if (config == null) {
            logExit("setupConfig(HashMap)");  //$NON-NLS-1$
            return;
        }
        String delimValue = (String) config.get("Delimiter");  //$NON-NLS-1$
        if (delimValue != null && delimValue.length() > 0) {
            if ("\\t".equalsIgnoreCase(delimValue) || "tab".equalsIgnoreCase(delimValue)) {  //$NON-NLS-1$  //$NON-NLS-2$
                charDelimiter = TAB_CHAR;
                delimiter = "\t";  //$NON-NLS-1$
                isCharDelimiter = true;
            }
            else
                if (delimValue.length() == 1) {
                    charDelimiter = delimValue.charAt(0);
                    delimiter = delimValue;
                    isCharDelimiter = true;
                }
                else {
                    delimiter = delimValue;
                }
        }

        String escapeCharacter = (String) config.get("Escape");  //$NON-NLS-1$
        if (escapeCharacter != null && escapeCharacter.length() > 0) {
            if (escapeCharacter.length() == 1) {
                charEscape = escapeCharacter.charAt(0);
                escape = escapeCharacter;
            }
        }
        String tns = (String)config.get("TargetNamespace");  //$NON-NLS-1$
        if (tns != null && tns.length() > 0) 
        {
            targetNamespace = tns;
        }
        logExit("setupConfig(config)");  //$NON-NLS-1$
    }


    /**
     * Adds a token value and the proper delimiter to the data String being
     * created (DOStringBuffer). Also contains the logic to check for the escape
     * and delimiter character in the token value passed in if needed.
     * 
     * @param checkDelimiter Determines whether or not logic to check token value for
     * delimiter and escape character is performed. If value is 'true', logic is performed. 
     * If value is 'false' logic is not performed.
     * @param tokenValue Token value to be added to the string being built.
     */
    protected void setSimpleToken(boolean checkDelimiter,
                                  String tokenValue)
    {
        logEntry("setSimpleToken(boolean, String)");  //$NON-NLS-1$
        if ("".equals(tokenValue)) {  //$NON-NLS-1$
            DOStringBuffer.append(" ");  //$NON-NLS-1$
            DOStringBuffer.append(delimiter);
            logExit("setSimpleToken(boolean, String)");  //$NON-NLS-1$
            return;
        }
        if (!checkDelimiter) {
            DOStringBuffer.append(tokenValue);
            DOStringBuffer.append(delimiter);
            logExit("setSimpleToken(boolean, String)");  //$NON-NLS-1$
            return;
        }
        if (checkDelimiter) {
            int delimiterIndex = 0;
            int escapeIndex = 0;
            int searchHeadIndex = 0;
            int tokenLength = tokenValue.length();
            do {
                delimiterIndex = tokenValue.indexOf(delimiter, searchHeadIndex);
                escapeIndex = tokenValue.indexOf(escape, searchHeadIndex);
                if (escapeIndex == -1) //escape charactor is not
                // present in pvTokenValue
                {
                    if (delimiterIndex > -1) //delimiter found
                    {
                        DOStringBuffer.append(tokenValue.substring(searchHeadIndex, delimiterIndex));
                        DOStringBuffer.append(escape);
                        DOStringBuffer.append(delimiter);
                        searchHeadIndex = delimiterIndex + delimiter.length();
                        //searchHeadIndex = delimiterIndex + 1;
                    }
                    else //both delimiter and escape are not found
                    {
                        DOStringBuffer.append(tokenValue.substring(searchHeadIndex, tokenLength));
                        DOStringBuffer.append(delimiter);
                        searchHeadIndex = tokenLength + 1;
                    }
                }
                else //escape is found
                {
                    if (delimiterIndex > -1) //delimiter found
                    {
                        if (escapeIndex >= delimiterIndex) {
                            //append the upto the one with smaller
                            // index first
                            DOStringBuffer.append(tokenValue.substring(searchHeadIndex, delimiterIndex));
                            DOStringBuffer.append(escape);
                            DOStringBuffer.append(delimiter);
                            searchHeadIndex = delimiterIndex + delimiter.length();
                            //searchHeadIndex = delimiterIndex + 1;
                        }
                        else {
                            DOStringBuffer.append(tokenValue.substring(searchHeadIndex, escapeIndex));
                            DOStringBuffer.append(escape);
                            DOStringBuffer.append(escape);
                            searchHeadIndex = escapeIndex + escape.length();
                            //searchHeadIndex = escapeIndex + 1;
                        }
                    }
                    else // if (delimiterIndex > -1 ) //delimiter NOT
                    // found
                    //only escape is found
                    {
                        DOStringBuffer.append(tokenValue.substring(searchHeadIndex, escapeIndex));
                        DOStringBuffer.append(escape);
                        DOStringBuffer.append(escape);
                        searchHeadIndex = escapeIndex + escape.length();
                        //searchHeadIndex = escapeIndex + 1;
                    }
                }
            } while (searchHeadIndex < tokenLength); //not in the
        }
        logExit("setSimpleToken(boolean, String)");  //$NON-NLS-1$
    } // end setSimpleToken


    /**
     * Performs search for char delimiter and returns string before the
     * delimter. Uses literal char comparisons for searching.
     * 
     * @return String value that is the next token that appers in the charBuffer.
     */
    protected String charDelimiterRead()
    {
        logEntry("charDelimiterRead()");  //$NON-NLS-1$
        StringBuffer buffer = new StringBuffer(BUFFERSIZE);

        int currentStartIndex = currentIndex;
        int endIndex = currentStartIndex;
        while (endIndex < charBuffer.length) {
            // Check if this character is a delimiter
            if (charBuffer[endIndex] == charDelimiter) {
                // check previous character to see if it is an escape character
                if (charBuffer[endIndex - 1] != charEscape) {
                    // it isn't so we found an unescaped delimiter and thus the
                    // end of our token
                    break;
                }
                // it is an escaped delimiter, so save the portion up to
                // the escaped character in our buffer
                buffer.append(charBuffer, currentStartIndex, endIndex - 1 - currentIndex);

                // append the delimiter in the buffer ( since it was an
                // escaped delimiter
                buffer.append(charDelimiter);

                endIndex++;
                currentStartIndex = endIndex;
            }
            // check if this character is an escape character and the previous
            // character is an escape character
            else
                if (charBuffer[endIndex] == charEscape && charBuffer[endIndex - 1] == charEscape) {
                    // we found an escaped escape character, so save the portion
                    // up to the first escape character in the buffer
                    buffer.append(charBuffer, currentStartIndex, endIndex - currentIndex);

                    // move the startIndex to the character past the second
                    // escape character
                    endIndex++;
                    currentStartIndex = endIndex;
                }
                else {
                    // this character is not a delimiter, and not an escaped
                    // escape
                    // character so move to the next character
                    endIndex++;
                }
        }

        currentIndex = endIndex + 1;
        buffer.append(charBuffer, currentStartIndex, endIndex - currentStartIndex);
        logExit("charDelimiterRead()");  //$NON-NLS-1$
        return buffer.toString();

    } // end charDelimiterRead


    /**
     * Performs search for string delimiter and returns string before the
     * delimter. Uses String.indexOf for searching.
     * 
     * @return String value that is the next token that appers in the
     *         charBuffer.
     * 
     *  
     */
    protected String stringDelimiterRead()
    {
        String lvResult = null;
        StringBuffer lvResultBuffer = new StringBuffer(attrValueBufferSize);

        int delimiterIndex = -1;
        int escapIndex = -1;

        if (!charSubstring(charBuffer, startIndex, startIndex + delimiter.length()).equalsIgnoreCase(delimiter)
            && startIndex < charBuffer.length) {
            // This condition ensures non-empty token
            do {
                // Check the data for the escape and delimiter characters
                escapIndex = charIndexOf(charBuffer, escape, startIndex);
                delimiterIndex = charIndexOf(charBuffer, delimiter, startIndex);

                if (escapIndex > -1 && escapIndex < delimiterIndex) {
                    // Checks the character that appears in the string just
                    // after
                    //  the escape character.
                    String lvCharBehindEscape = charSubstring(charBuffer, escapIndex + 1, escapIndex + 1
                                                                                          + delimiter.length());

                    if (lvCharBehindEscape.equalsIgnoreCase(escape)) {
                        // Append the data up until the escape character
                        // appears, but
                        //  skip the next character.
                        lvResultBuffer.append(charSubstring(charBuffer, startIndex, escapIndex));
                        lvResultBuffer.append(charSubstring(charBuffer, escapIndex + 1, escapIndex + 1
                                                                                        + escape.length()));
                    }
                    else
                        if (lvCharBehindEscape.equalsIgnoreCase(delimiter)) {
                            lvResultBuffer.append(charSubstring(charBuffer, startIndex, escapIndex));
                            lvResultBuffer.append(charSubstring(charBuffer, escapIndex + 1, escapIndex + 1
                                                                                            + delimiter.length()));
                        }
                        else {
                            // This is the case that the escape character is
                            // part of the
                            //  actual data and is not being used as an escape
                            // character.
                            lvResultBuffer.append(charSubstring(charBuffer, startIndex, escapIndex + escape.length()
                                                                                        + 1));
                        }
                    startIndex = escapIndex + escape.length() + 1;//ensured
                }
                else {
                    if (delimiterIndex > -1) {
                        lvResultBuffer.append(charSubstring(charBuffer, startIndex, delimiterIndex));
                    }
                    else {
                        lvResultBuffer.append(charSubstring(charBuffer, startIndex, charBuffer.length));
                    }

                    // Increment the start index to skip over the delimiter
                    startIndex = delimiterIndex + delimiter.length();
                }

            }

            while (escapIndex > -1 && escapIndex < delimiterIndex && startIndex < charBuffer.length);
            lvResult = lvResultBuffer.toString();
            if (startIndex >= charBuffer.length) {
                endOfDOString = true;
            }
            return lvResult;
        }
        if (charSubstring(charBuffer, startIndex, startIndex + delimiter.length()).equalsIgnoreCase(delimiter)
            && startIndex < charBuffer.length) {
            // Define new starting search point for next token
            startIndex = startIndex + delimiter.length();
            return lvResult;
        }
        endOfDOString = true;
        return lvResult;

    } // end stringDelimiterRead


    /**
     * Returns a new StringBuffer that is a substring of charArray. The
     * substring begins with the character at the specified begin index and
     * extends to the end index. Examples:
     * 
     * charArray = {'d', 'o', 'o', 'r'}; charSubstring(charArray, 1, 2) returns
     * "oo"
     * 
     * @param charArray -
     *            the array to use for substring.
     * @param begin -
     *            the beginning index, inclusive.
     * @param end -
     *            the ending index, non-inclusive.
     * @return the specified substring.
     */
    protected String charSubstring(char[] charArray,
                                   int begin,
                                   int end)
    {
        StringBuffer returnString = new StringBuffer();

        for (int i = begin; i < end; i++) {
            returnString.append(charArray[i]);
        }

        return returnString.toString();
    } // end charSubstring


    /**
     * Returns the index within this charArray of the first occurrence of the
     * specified substring, starting at the specified begin index.
     * 
     * There is no restriction on the value of <code>fromIndex</code>. If it
     * is negative, it has the same effect as if it were zero: this entire
     * string may be searched. If it is greater than the length of this string,
     * it has the same effect as if it were equal to the length of this string:
     * -1 is returned.
     * 
     * @param charArray -
     *            the charArray to search in.
     * @param pattern -
     *            the string to search for.
     * @param begin -
     *            the index to start the search from.
     * @return If the string argument occurs as a substring within this object
     *         at a starting index no smaller than <code>fromIndex</code>,
     *         then the index of the first character of the first such substring
     *         is returned. If it does not occur as a substring starting at
     *         <code>fromIndex</code> or beyond, -1 is returned.
     */
    protected static int charIndexOf(char[] charArray,
                                     String pattern,
                                     int begin)
    {
    	if(pattern == null)
    	{    	
    		return 0;
    	}
        char charPattern[] = pattern.toCharArray();
        int max = charArray.length - pattern.length();

        if (begin >= charArray.length) {
            if (charArray.length == 0 && begin == 0 && pattern.length() == 0) {
                /* There is an empty string at index 0 in an empty string. */
                return 0;
            }
            /* Note: begin index might be near -1>>>1 */
            return -1;
        }
        if (begin < 0) {
            begin = 0;
        }
        if (pattern.length() == 0) {
            return begin;
        }

        char first = charPattern[0];
        int beginIndex = begin;

        while (true) {
            /* Look for first character. */
            while (beginIndex <= max && charArray[beginIndex] != first) {
                beginIndex++;
            }
            if (beginIndex > max) {
                return -1;
            }

            /* Found first character, now look at the rest of v2 */
            int j = beginIndex + 1;
            int end = j + pattern.length() - 1;
            int k = 1;
            while (j < end) {
                if (charArray[j++] != charPattern[k++]) {
                    beginIndex++;
                    /* Look for str's first char again. */
                    j = end;
                }
            }
            return beginIndex; /* Found whole string. */
        }
    } // end charIndexOf


    private void logEntry(String message)
    {
        if (LOGGER.isLoggable(Level.INFO)) {
            LOGGER.entering(CLASSNAME, message);
        }
    }


    private void logExit(String message)
    {
        if (LOGGER.isLoggable(Level.INFO)) {
            LOGGER.exiting(CLASSNAME, message);
        }
    }

}
