/*BEGIN OCO COPYRIGHT
 *************************************************************************
 *
 * IBM Confidential
 * OCO Source Materials
 * 5724-L01, 5655-N53, 5724-I82, 5655-R15
 * (C) Copyright IBM Corporation 2006.
 * The source code for this program is not published or otherwise
 * divested of its trade secrets, irrespective of what has been
 * deposited with the U.S. Copyright Office.
 *
 **************************************************************************
 *END OCO COPYRIGHT
 */

package com.ibm.websphere.samples.databinding.util;


import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map.Entry;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.sdo.EProperty;
import org.eclipse.emf.ecore.util.EcoreUtil;

import com.ibm.websphere.bo.BOFactory;
import com.ibm.websphere.samples.databinding.exceptions.NameValueHandlerException;
import com.ibm.websphere.samples.databinding.log.HandlerUtil;
import com.ibm.websphere.sca.ServiceManager;
import commonj.sdo.DataObject;
import commonj.sdo.Property;
import commonj.sdo.Type;

/**
 * The NameValueDataHandler converts a name value stream of data to a 
 * data object and vice versa. 
 * A name value stream for a customer business object which contain 
 * properties
 * id, firstName, lastName, contained object address with id looks as 
 * follows 
 * BusinessObject = Customer
    id = 23
    firstName = John
    lastName = Doe
    address = 1
        BusinessObject = Address
        id = 10
        street = 577 Airport Blvd
        city = Burlingame
        state = CA
        zip = 94010
        end = Address
    end = Customer
 * There are configuration properties like unset, blank,
 * blankValue, encoding, targetNamespace that can be
 * passed to the handler.
 * <p>
 * @ibm-api
 */

public class NameValueDataHandler
{
    private static final String BOTOKEN = "BusinessObject";  //$NON-NLS-1$

    private boolean skipUnset;
    private String unset;
    private String blank;
    private String blankValue;
    private String encoding;
    private String targetNamespace;

    //for the BOString parsing
    private int lineNum; //for error messaging
    private int lineEndIndex;
    private int boStringLength;
    private int depth;
    private StringBuffer boStringBuffer;
    private String boString = null;

    private static ServiceManager serviceManager = new ServiceManager();
    private static final BOFactory BOFACTORY = (BOFactory) serviceManager.locateService("com/ibm/websphere/bo/BOFactory");  //$NON-NLS-1$

    private static final String NEWLINE = System.getProperty("line.separator");  //$NON-NLS-1$
    private static final String CLASSNAME = DelimitedDataHandler.class.getName();
    private static final Logger LOGGER = HandlerUtil.initLogger(CLASSNAME);

//	BEGIN COPYRIGHT VARIABLE
	public static final String COPYRIGHT = "(C) Copyright IBM Corporation 2006.";	//$NON-NLS-$1
	// END COPYRIGHT VARIABLE

    /**
     * The constructor initializes all the fields in the class
     */
    public NameValueDataHandler()
    {
        lineNum = 0;
        lineEndIndex = 0;
        boStringLength = 0;
        depth = 0;		
        if (boStringBuffer != null) {
            int len = boStringBuffer.length();
            boStringBuffer.delete(0, len);
        }
        boStringBuffer = new StringBuffer();
        
        //configuration properties
        skipUnset = false;
        unset = new String("unset");  //$NON-NLS-1$
        blank = new String("");  //$NON-NLS-1$
        blankValue = new String("");  //$NON-NLS-1$
        encoding = null;
        targetNamespace = new String("http://www.ibm.com/crm");  //$NON-NLS-1$
    }


    /**
     * Override the properties of the handler with the ones passed in the
     * config object.
     * @param config HashMap of properties of handler
     */
    protected void setupConfig(HashMap config)
    {
        if (config == null || config.isEmpty()) {
            return;
        }

        Iterator iterator = config.entrySet().iterator();
        while (iterator.hasNext()) {
            Entry entry = (Entry) iterator.next();
            if (entry.getKey().equals("skipUnset")) {  //$NON-NLS-1$
                skipUnset = ((Boolean) (entry.getValue())).booleanValue();
                continue;
            }
            if (entry.getKey().equals("unset")) {  //$NON-NLS-1$
                unset = (String) entry.getValue();
                continue;
            }
            if (entry.getKey().equals("blank")) {  //$NON-NLS-1$
                blank = (String) entry.getValue();
                continue;
            }
            if (entry.getKey().equals("blankValue")) {  //$NON-NLS-1$
                blankValue = (String) entry.getValue();
                continue;
            }
            if (entry.getKey().equals("encoding")) {  //$NON-NLS-1$
                encoding = (String) entry.getValue();
                continue;
            }
            if (entry.getKey().equals("TargetNamespace")) {  //$NON-NLS-1$
                targetNamespace = (String) entry.getValue();
                continue;
            }

        }
    }


    /**
     * Converts the input bytes to a data object. The input bytes 
     * are in name value format as follows
     * BusinessObject = SAPCustomer
     * firstName = Zeenat
     * address = 1
     *     BusinessObject = SAPAddress
     *     id = unset
     *     phone = 2
     *         BusinessObject = SAPPhone
     *         id = 13
     *         end = SAPPhone
     *         BusinessObject = SAPPhone
     *         id = 15
     *         end = SAPPhone
     *     end = SAPAddress
     * code = 2
     * code = 10
     * code = 20
     * address1 = unset
     * code1 = unset
     * end = SAPCustomer
     * @param bytes  input bytes to be converted to DataObject.
     * @param config properties on the handler
     * @return the data object converted from the bytes
     */
    public DataObject getDataObject(byte[] bytes,
                                    HashMap config)
    	throws NameValueHandlerException
    {
        logEntry("getDataObject(byte[], HashMap");  //$NON-NLS-1$
        HandlerUtil.logConfig(LOGGER, CLASSNAME, "getDataObject(byte[], HashMap)", config);  //$NON-NLS-1$

        setupConfig(config);
        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
        try {
            outputStream.write(bytes);
        } catch (IOException e) {
            throw new NameValueHandlerException("Name Value Handler error when reading the serialized data", e);  //$NON-NLS-1$
        }
        boString = outputStream.toString();
        boStringLength = boString.length();
        DataObject responseObj = parseDataObjectString();

        HandlerUtil.logDataObject(LOGGER, CLASSNAME, "getDataObject(byte[], HashMap)", //$NON-NLS-1$
                                  responseObj);
        logExit("getDataObject(byte[], HashMap");  //$NON-NLS-1$
        return responseObj;
    }


    /**
     *   Populates a given DataObject with data converted
     *   from String object. The String object is in the
     *   NameValue format. Allows configuration options to
     *   be set by passing in a hashmap which contains the configuration
     *   properties.
     *
     *   @param  serializedData  String which contains the data to be
     *                           converted to a business object.
     *   @param  dataObject          DataObject to be populated with
     *                           data serializedData.
     *   @param  config          Contains the configuration properties
     *                           on the handler.
     * @return DataObject
     */
    public DataObject getDataObject(String serializedData,
                              DataObject dataObject,
                              HashMap config)
    	throws NameValueHandlerException
    {
        logEntry("getDataObject(String, DataObject, HashMap");  //$NON-NLS-1$
        HandlerUtil.logConfig(LOGGER, CLASSNAME, "getDataObject(byte[], HashMap)", config);  //$NON-NLS-1$

        setupConfig(config);
        boString = serializedData;
        boStringLength = boString.length();
        DataObject responseObj = parseDataObjectString(dataObject);

        HandlerUtil.logDataObject(LOGGER, CLASSNAME, "getDataObject(String, DataObject, HashMap)", //$NON-NLS-1$
                                  responseObj);
        logExit("getDataObject(String, DataObject, HashMap");  //$NON-NLS-1$
        return responseObj;
    }


    /**
     *   Converts the incoming String to the corresponding data object.
     *   In this case, it is assumed that the string contains the name of 
     *   the data object. The String object is in the NameValue format. 
     *   Allows configuration options to be set by passing in a hashmap
     *   which contains the configuration properties.
     *   @param  serializedData  String which contains the data to be
     *                           converted to a business object.
     *   @param  config          Contains the configuration properties
     *                           on the handler.
     *   @return DataObject converted from the incoming string.
     */

    public DataObject getDataObject(String serializedData,
                                    HashMap config)
    	throws NameValueHandlerException
    {
        logEntry("getDataObject(String, HashMap");  //$NON-NLS-1$
        HandlerUtil.logConfig(LOGGER, CLASSNAME, "getDataObject(byte[], HashMap)", config);  //$NON-NLS-1$

        setupConfig(config);
        boString = serializedData;
        boStringLength = boString.length();
        DataObject responseObj = parseDataObjectString();
        
        HandlerUtil.logDataObject(LOGGER, CLASSNAME, "getDataObject(String, HashMap)", //$NON-NLS-1$
                                  responseObj);
        logExit("getDataObject(String, HashMap");  //$NON-NLS-1$
        return responseObj;
    }


    /**
     *   Creates a String based on the data contained in a DataObject.
     *   Allows configuration options to be set by passing in a hashmap
     *   which contains the config information.
     *   The output String is in the NameValue format.
     *   @param  theObj      DataObject which contains the data to
     *                       be included in the String.
     *   @param  config      Configuration propery values for the handler.
     *   @return String       which contains the data from theObj in name value
     *               pair format.
     */
    public String getStringFromDataObject(DataObject theObj,
                                          HashMap config)
    	throws NameValueHandlerException
    {
        logEntry("getStringFromDataObject(DataObject, HashMap");  //$NON-NLS-1$
        HandlerUtil.logConfig(LOGGER, CLASSNAME, "getStringFromDataObject(DataObject, HashMap)", config);  //$NON-NLS-1$

        setupConfig(config);
        serializeDataObject(theObj, true);
        String dataObjectString = boStringBuffer.toString();
        
        LOGGER.logp(Level.FINER, CLASSNAME, "getStringFromDataObject(DataObject, HashMap)", //$NON-NLS-1$
                    "Created the Stream : \"{0}\"", dataObjectString);  //$NON-NLS-1$
        logExit("getStringFromDataObject(DataObject, HashMap");  //$NON-NLS-1$
        
        return dataObjectString;
    }


    /**
     *   Creates a byte array based on the data contained in a DataObject.
     *   Allows properties to be set on the handler.
     *   The output bytes are in the NameValue format.
     *   @param  dataObject      DataObject which contains the data to
     *                       be included in the String.
     *   @param  config      Configuration propery values for the handler.
     *   @return     String which contains the data from theObj in name value
     *               pair format.
     */
    public byte[] getBytesFromDataObject(DataObject dataObject,
                                         HashMap config)
    	throws NameValueHandlerException
    {
        logEntry("getBytesFromDataObject(DataObject, HashMap)");  //$NON-NLS-1$
        HandlerUtil.logDataObject(LOGGER, CLASSNAME, "getStringFromDataObject(DataObject, HashMap)", dataObject);  //$NON-NLS-1$
        HandlerUtil.logConfig(LOGGER, CLASSNAME, "getStringFromDataObject(DataObject, HashMap)", config);  //$NON-NLS-1$

        setupConfig(config);
        String dataObjectString = getStringFromDataObject(dataObject, config);        

        try {
            if (encoding != null) {
                logExit("getBytesFromDataObject(DataObject, HashMap)");  //$NON-NLS-1$
                return dataObjectString.getBytes(encoding);
            }
            logExit("getBytesFromDataObject(DataObject, HashMap)");  //$NON-NLS-1$
            return dataObjectString.getBytes();
        } catch (UnsupportedEncodingException e) {
            String message = "The encoding type " + encoding + " is not supported";  //$NON-NLS-2$  //$NON-NLS-1$
            LOGGER.logp(Level.SEVERE, CLASSNAME, "getBytesFromDataObject(DataObject, HashMap)", message);  //$NON-NLS-1$
            throw new NameValueHandlerException(message, e);
        }
    }


    /**
     *   Build a string with the data contained in the DataObject.
     *   Operates on mBOStringBuffer.
     *   @param  dataObject        DataObject to be serialized.
     *   @param  mainBO  Boolean flag to indicate whether this object is the top
     *                       level business object.  Level of BO used to properly
     *                       format the output (indent correctly).
     */
    protected void serializeDataObject(DataObject dataObject,
                                       boolean mainBO)
    	throws NameValueHandlerException
    {
        logEntry("serializeDataObject(DataObject, boolean");  //$NON-NLS-1$
        // mDepth is used to determine indention of the name value pairs
        if (!mainBO) {
            ++depth;
        }

        Type type = dataObject.getType();
        String name = type.getName();

        try {
            this.setSimpleToken(BOTOKEN, name);
            this.serializeAllProperties(dataObject);
            this.setSimpleToken("end", name);  //$NON-NLS-1$
            --depth;
        } catch (Exception e) {
            String message = "Error serializing BO " + name;  //$NON-NLS-1$
            LOGGER.logp(Level.SEVERE, CLASSNAME, "serializeDataObject(DataObject, boolean)", message);  //$NON-NLS-1$            
            throw new NameValueHandlerException(message, e);
        }
        
        logExit("serializeDataObject(DataObject, boolean");  //$NON-NLS-1$
    }

    /**
     * First serialize the size of the list e.g. address = 2, next serialize the
     * name of the property e.g. BusinessObject = Address
     * corresponding to the child data object from the list and serialize its
     * property values to a name value string.
     *
     * @param parentDataObject
     * @param propertyName 
     */
    protected void serializeChildDataObjectList(DataObject parentDataObject,
                                                String propertyName)
    	throws NameValueHandlerException
    {
        logEntry("serializeChildDataObjectList(DataObject, Property)");  //$NON-NLS-1$        
        LOGGER.logp(Level.FINE, CLASSNAME, "serializeChildDataObjectList(DataObject, Property)",  //$NON-NLS-1$
                    "Serializing complex type list property \"{0}\"", propertyName);  //$NON-NLS-1$
        
        List propertyList = parentDataObject.getList(propertyName);

        // If the container is empty, add unset is the value of the
        //  attribute.
        if (propertyList.isEmpty() && !skipUnset) {
            this.setSimpleToken(propertyName, unset);
            return;
        }
        if (!propertyList.isEmpty()) {
            // Add the attribute name and count to the serialized data
            int objectCount = propertyList.size();
            this.setSimpleToken(propertyName, new Integer(objectCount));

            // Loop through the container and add each of the contained
            //  objects to the serialized data
            for (int index = 0; index < objectCount; ++index) {
                serializeDataObject((DataObject) propertyList.get(index), false);
            }
        }
        logExit("serializeChildDataObjectList(DataObject, Property)");  //$NON-NLS-1$
    }

    /**
     * Serialize the property values of child data object to a name value string
     *
     * @param parentDataObject
     * @param propertyName 
     */
    protected void serializeChildDataObject(DataObject parentDataObject,
                                            String propertyName)
    	throws NameValueHandlerException
    {
        logEntry("serializeChildDataObject(DataObject, String)");  //$NON-NLS-1$
        LOGGER.logp(Level.FINE, CLASSNAME, "serializeChildDataObject(DataObject, String)",  //$NON-NLS-1$
                    "Serializing complex type property \"{0}\"", propertyName);  //$NON-NLS-1$
        
        DataObject childDataObject = parentDataObject.getDataObject(propertyName);

        // If the container is empty, add unset as the value of the
        //  attribute.
        if (childDataObject == null && !skipUnset) {
            this.setSimpleToken(propertyName, unset);
            return;
        }
        if (childDataObject != null) {
            this.setSimpleToken(propertyName, "1");  //$NON-NLS-1$
            // Add the contained object to the serialized data
            serializeDataObject(childDataObject, false);
        }
        
        logExit("serializeChildDataObject(DataObject, String)");  //$NON-NLS-1$
    }


    /**
     *   Adds the name value pairs for the attributes to the serialized data object.
     *   @param dataObject     Data object which contains the data to be serialized.
     */
    protected void serializeAllProperties(DataObject dataObject)
    	throws NameValueHandlerException
    {
        logEntry("serializeAllProperties(DataObject)");  //$NON-NLS-1$
        
        Type type = dataObject.getType();
        Iterator propertyIterator = type.getProperties().iterator();

        while (propertyIterator.hasNext()) {
            Property property = (Property) propertyIterator.next();
            String propertyName = property.getName();
            // Check if the attribute is a simple attribute or a container
            if (property.isContainment() && property.isMany()) {
                serializeChildDataObjectList(dataObject, propertyName);
            }
            if (property.isContainment() && property.isMany() == false) {
                // Single cardinality 
                serializeChildDataObject(dataObject, propertyName);
            }
            if (property.isContainment() == false && property.isMany() == false) {
                // Add the attribute name and value to the serialized data
                if (dataObject.isSet(propertyName)) {
                    String propertyValue = dataObject.getString(propertyName);
                    setSimpleToken(propertyName, propertyValue);
                }
                else {
                    setSimpleToken(propertyName, unset);
                }
            }
            if (property.isContainment() == false && property.isMany()) {
                // this is a simple type list
                serializeSimpleTypeList(dataObject, property);
            }
        }
        logExit("serializeAllProperties(DataObject)");  //$NON-NLS-1$
    }

    /**
     * Serialize out the size of the list and then the values in the list.
     * 
     * @param dataObject The DataObject that contains the list of simple type
     * @param property The property that corresponds to the list of simple type
     */
    protected void serializeSimpleTypeList(DataObject dataObject,
                                           Property property)
    {
        logEntry("serializeSimpleTypeList(DataObject, Property)");  //$NON-NLS-1$
        
        String propertyName = property.getName();
        List list = dataObject.getList(propertyName);

        //list is null or empty, so add unset and return
        if (list == null || list.isEmpty()) {
            this.setSimpleToken(propertyName, null);
            return;
        }

        // Add list size to the fixed width String
        int listSize = list.size();
        this.setSimpleToken(propertyName, Integer.toString(listSize));


        Iterator listIterator = list.iterator();
        depth++;
        while (listIterator.hasNext()) {
            Object listValue = listIterator.next();
            String propertyValue = convertToString(property, listValue);
            this.setSimpleToken(propertyName, propertyValue);
        }
        depth--;
        logExit("serializeSimpleTypeList(DataObject, Property)");  //$NON-NLS-1$
    }

    /**
     * Convert the value of any type to a string type so it can be tokenized to the stream
     *
     * @param property Property whose value is to be converted to string
     * @param objValue value to be converted to string
     * @return String
     */
    protected String convertToString(Property property,
                                     Object objValue)
    {
        EDataType dataType = (EDataType) ((EProperty) property).getEStructuralFeature().getEType();
        String value = EcoreUtil.convertToString(dataType, objValue);
        return value;
    }


    /**
     *   Adds a single name value pair to mBOStringBuffer based on the token name and
     *   passed in as parameters.
     *   @param  tokenName    This is the name of the token to be added (i.e. right
     *                           side of the equals sign).
     *   @param  tokenValue    This is the value of the token to be added (i.e. left
     *                           side of the equals sign).
     */
    protected void setSimpleToken(String tokenName,
                                  Object tokenValue)
    {
        logEntry("setSimpleToken(String, Object)");  //$NON-NLS-1$
        if (tokenValue != null) //write as given
        {
            this.indent();
            boStringBuffer.append(tokenName);
            boStringBuffer.append(" = ");  //$NON-NLS-1$
            if (tokenValue.equals(blankValue)) { //CxBlank
                boStringBuffer.append(blank);
            
            } else {
                boStringBuffer.append(tokenValue.toString());
            }
            boStringBuffer.append(NEWLINE);
        }
        else
            if (!skipUnset) {
                this.indent();
                boStringBuffer.append(tokenName);
                boStringBuffer.append(" = ");  //$NON-NLS-1$
                boStringBuffer.append(unset);
                boStringBuffer.append(NEWLINE);
            }
        logExit("setSimpleToken(String, Object)");  //$NON-NLS-1$
    }


    /**
     *   Adds as many tabs as the degree of child object depth to
     *   mBOStringBuffer.  Number of tabs determined by member
     *   variable mDepth.
     */
    protected void indent()
    {
        for (int tabCount = 0; tabCount < depth; tabCount++) {
            boStringBuffer.append("\t");  //$NON-NLS-1$
        }
    }

    /**
     *   Parses the serialized data object represenation.  
     *   Assumes that the serialized representation has as the
     *   first line the data object name token.

     *   @return     DataObject with attributs populated based on
     *               data in the serialized business object.
     */
    protected DataObject parseDataObjectString()
    	throws NameValueHandlerException
    {
        logEntry("parseDataObjectString()");  //$NON-NLS-1$
        String[] tokenPair = getNextToken();
        if (BOTOKEN.compareTo(tokenPair[0]) != 0) {
            throw new NameValueHandlerException("Name Value Data Handler expects a Business Object Name");  //$NON-NLS-1$
        }

        String parsedDataObjectName = tokenPair[1];
        DataObject dataObject = BOFACTORY.create(targetNamespace, parsedDataObjectName);
        LOGGER.logp(Level.FINER, CLASSNAME, "getDataObjectFromString(String, HashMap)", //$NON-NLS-1$
                    "Created a DataObject: \"{0}\"", targetNamespace + "\\" + parsedDataObjectName);  //$NON-NLS-1$            
    //  Parse through the BO attributes in the
        parseAllProperties(dataObject);
        HandlerUtil.logDataObject(LOGGER, CLASSNAME, "getDataObject(String, HashMap)", dataObject);  //$NON-NLS-1$        
        logExit("parseDataObjectString()");  //$NON-NLS-1$
        return dataObject;
    }


    /**
     *   Parses the serialized data object represenation.  Checks the
     *   type against what is contianed in the serialized data if a type is
     *   passed in.  Assumes that the serialized representation has as the
     *   first line the data object name token.
     *   @param  dataObject    The name of the data object which should be
     *                       created from the current position in the serialized
     *                       data.  Assume it is 'null' only for the top level
     *                       data object, otherwise populated.
     *   @return     DataObject with attributs populated based on
     *               data in the serialized business object.
     */
    protected DataObject parseDataObjectString(DataObject dataObject)
    	throws NameValueHandlerException
    {
        logEntry("parseDataObjectString(DataObject)");  //$NON-NLS-1$
        String dataObjectName = dataObject.getType().getName();
        dataObject = BOFACTORY.create("http://Test/com/wbi", dataObjectName);  //$NON-NLS-1$
        //  Parse through the BO attributes in the
        this.parseAllProperties(dataObject);
        logExit("parseDataObjectString(DataObject)");  //$NON-NLS-1$
        return dataObject;
    }


    /**
     *   Parses the list of attributes and data in mBOString to populate the attributes
     *   of the DataObject passed in as a parameter.
     *   Assumes data includes token for ObjectEventId attribute.
     *   @param  dataObject    DataObject which should be populated based on the
     *                   name value pairs in the serialized data.
     */
    protected void parseAllProperties(DataObject dataObject)
    	throws NameValueHandlerException
    {
        logEntry("parseAllProperties(DataObject)");  //$NON-NLS-1$
        if (dataObject == null) {
            throw new NameValueHandlerException(" Name Value Data Handler found unexpected ending of BO "  //$NON-NLS-1$
                                                + ".  Please check your data format; ");  //$NON-NLS-1$
        }

        Type type = dataObject.getType();
        String dataObjectName = type.getName();

        String[] tokenPair = getNextToken();
        String parsedPropertyName = tokenPair[0];
        String parsedPropertyValue = tokenPair[1];

        while (!"end".equals(parsedPropertyName)) {  //$NON-NLS-1$
            // Set the attribute name and value locally based on the current token
            //  information.

            Property property = null;
            try {
                property = type.getProperty(parsedPropertyName);
            } catch (Exception e) {
                throw new NameValueHandlerException("The property " + parsedPropertyName  //$NON-NLS-1$
                                                    + " does not exist in the DataObject " + dataObjectName, e);  //$NON-NLS-1$
            }

            
            if (property.isContainment() && property.isMany()) {
                parseChildDataObjectList(dataObject, property, parsedPropertyValue);
            }
            if (property.isContainment() && property.isMany() == false) {
                parseChildDataObject(dataObject, property, parsedPropertyValue);
            }
            if (property.isContainment() == false && property.isMany()) {
                parseSimpleTypeList(dataObject, property, parsedPropertyValue);
            }
            if (property.isContainment() == false && property.isMany() == false) {
                LOGGER.logp(Level.FINE, CLASSNAME, "parseSimpleTypeList(DataObject, Property, String)",  //$NON-NLS-1$
                            "Parsing simple type property : \"{0}\"", property.getName());  //$NON-NLS-1$
                if (!parsedPropertyValue.equals(unset)) {
                    dataObject.setString(parsedPropertyName, parsedPropertyValue);
                
                } else {
                    dataObject.unset(parsedPropertyName);
                }
            }
            tokenPair = getNextToken();
            parsedPropertyName = tokenPair[0];
            parsedPropertyValue = tokenPair[1];

        }
        logExit("parseAllProperties(DataObject)");  //$NON-NLS-1$
    }


    /**
     *   Populates a single cardinatlity child data object with the token
     *   values found in the name value string.
     *
     *   @param  parentDataObject        Parent business object.
     *   @param  propertyValue     Number of instances of the child
     *                             object as parsed from the name value string.
     * @param property 
     */
    protected void parseChildDataObject(DataObject parentDataObject,
                                        Property property,
                                        String propertyValue)
    	throws NameValueHandlerException
    {
        logEntry("parseChildDataObject(DataObject, Property, String)");  //$NON-NLS-1$
        
        if (propertyValue.equals(unset)) {
            return;
        }
        int objectCount = Integer.parseInt(propertyValue);

        if (objectCount == 0) {
            return;
        }
        if (objectCount < 0) {
            throw new NameValueHandlerException("The number of objects cannot be less than 0");  //$NON-NLS-1$
        }
        
        String propertyName = property.getName();
        LOGGER.logp(Level.FINE, CLASSNAME, "parseChildDataObject(DataObject, String)",  //$NON-NLS-1$
                    "Parsing child data object named \"{0}\"", propertyName);  //$NON-NLS-1$
        String[] tokenPair = getNextToken();
        String propertyType = property.getType().getName();

        if (propertyType.compareTo(tokenPair[1]) != 0)
        {
            throw new NameValueHandlerException("The type in the file " + tokenPair[0]  //$NON-NLS-1$
                                                + "does not match the type of the " + "property " + propertyType);  //$NON-NLS-1$  //$NON-NLS-2$
        }

        DataObject childDataObject = parentDataObject.createDataObject(propertyName);
        parseAllProperties(childDataObject);
        
        logExit("parseChildDataObject(DataObject, Property, String)");  //$NON-NLS-1$
    }


    /**
     *   Populates a child container based on token values in the delimited
     *   string.
     *   Expects all objects in the child container to be of the same type.
     *
     *   @param  parentDataObject        Parent data object.
     *   @param  propertyValue     Number of instances of the child
     *                                   object as parsed from the name value
     *                                   string.
     * @param property 
     */
    protected void parseChildDataObjectList(DataObject parentDataObject,
                                            Property property,
                                            String propertyValue)
    	throws NameValueHandlerException
    {
        logEntry("parseChildDataObjectList(DataObject, Property, String)");  //$NON-NLS-1$
        
        if (propertyValue.equals(unset)) {
            return;
        }
        int objectCount = Integer.parseInt(propertyValue);

        if (objectCount == 0) {
            return;
        }
        if (objectCount < 0) {
            throw new NameValueHandlerException("The number of objects cannot be less than 0");  //$NON-NLS-1$
        }
        String propertyName = property.getName();
        
        LOGGER.logp(Level.FINE, CLASSNAME, "parseChildDataObject(DataObject, String)",  //$NON-NLS-1$
                    "Parsing child data object \"{0}\"", propertyName);  //$NON-NLS-1$
        
        String propertyType = property.getType().getName();
        for (int index = 0; index < objectCount; index++) {
            String[] tokenPair = getNextToken();            

            if (propertyType.compareTo(tokenPair[1]) != 0) {
                throw new NameValueHandlerException("The type in the file " + tokenPair[0]  //$NON-NLS-1$
                                                    + "does not match the type of the " + "property " + propertyType);  //$NON-NLS-1$  //$NON-NLS-2$
            }

            DataObject childDataObject = parentDataObject.createDataObject(propertyName);
            parseAllProperties(childDataObject);
        }
        logExit("parseChildDataObjectList(DataObject, Property, String)");  //$NON-NLS-1$
    }

    /**
     * Populates a list of simple types with the token values found in the fixed
     * width string. 
     * 
     * @param dataObject    Parent data object. 
     * @param property      Property corresponding to the simple type list 
     * @param propertyValue size of the simple type list.
     */
    protected void parseSimpleTypeList(DataObject dataObject,
                                       Property property,
                                       String propertyValue)
    	throws NameValueHandlerException
    {
        logEntry("parseSimpleTypeList(DataObject, Property, String)");  //$NON-NLS-1$
        if (propertyValue.equals(unset)) {
            //TODO:trace
            return;
        }
        int listSize = Integer.parseInt(propertyValue);

        if (listSize == 0) {
            //TODO:trace
            return;
        }
        
        if (listSize < 0) 
        {
            throw new NameValueHandlerException("The list size is less than 0");  //$NON-NLS-1$
        }
        
        String propertyName = property.getName();
        LOGGER.logp(Level.FINE, CLASSNAME, "parseSimpleTypeList(DataObject, Property, String)",  //$NON-NLS-1$
                    "Parsing simple type list \"{0}\"", propertyName);  //$NON-NLS-1$
        
        List list = dataObject.getList(propertyName);
        for (int i = 0; i < listSize; i++) {
            //Get the next token based on the MaxLength of the attribute                      
            String[] tokenPair = this.getNextToken();

            if (tokenPair[1] == null) {
                throw new NameValueHandlerException("Encountered a null in the list");  //$NON-NLS-1$
            }
            if (tokenPair[1].equals(unset) || "".equals(tokenPair[1])) {  //$NON-NLS-1$
                throw new NameValueHandlerException("Encountered an unset in the list");  //$NON-NLS-1$
            }
            if (tokenPair[1].equals(blank) || " ".equals(tokenPair[1])) {  //$NON-NLS-1$
                throw new NameValueHandlerException("Encountered a blank in the list");  //$NON-NLS-1$
            }

            EStructuralFeature eFeature = ((EProperty) property).getEStructuralFeature();

            EDataType dataType = (EDataType) eFeature.getEType();
            Object value = EcoreUtil.createFromString(dataType, tokenPair[1]);
            list.add(value);
        }
        logExit("parseSimpleTypeList(DataObject, Property, String)");  //$NON-NLS-1$
    }


    /**
     *   Method to read the name and value of the next token in the serialized
     *   name value data.
     *   @returns        An array of strings with exactly two elements, the first
     *                   being the name of the pair, and the second the value.
     */
    protected String[] getNextToken()
    {
        String[] tokenPair = new String[2];

        // This moves the current location in the serialized BO string to the
        //  next line.
        try {
            tokenPair = getNextLine();
        } catch (Exception e) {
            tokenPair[0] = null;
            tokenPair[1] = null;
        }
        return tokenPair;
    }


    /** start and end delimiting index of line
     *  Expect mBOString in form of delimited by '\n' and empty
     *  line is not skipped.  But line number is still properly
     *  counted.
     *   @returns        An array of strings with exactly two elements, the first
     *                   being the name of the pair, and the second the value.
     */
    protected String[] getNextLine()
    	throws NameValueHandlerException
    {
        boolean isLineEmpty = true;
        int startIndex = 0;
        int endIndex = 0;

        logEntry("getNextLine()");  //$NON-NLS-1$
        if (lineEndIndex >= boStringLength) {
            throw new NameValueHandlerException("NameValue Handler encoutnered a premature end of serialized data");  //$NON-NLS-1$
        }

        int lineStartIndex = 0;
        while (isLineEmpty) {
            lineStartIndex = lineEndIndex;
            lineEndIndex = boString.indexOf(NEWLINE, lineStartIndex);
            ++lineNum;

            if (lineEndIndex > lineStartIndex) //at the end, the two will be equal
            {
                if (!this.emptyLine(lineStartIndex, lineEndIndex, startIndex)) {
                    isLineEmpty = false;
                }
            }
            lineEndIndex++; //skip over the '\n' character
        }

        startIndex = lineStartIndex;
        endIndex = lineEndIndex;

        while (Character.isWhitespace(boString.charAt(startIndex)) && startIndex < lineEndIndex) {
            startIndex++;
        }

        //trim right
        while (Character.isWhitespace(boString.charAt(endIndex - 1)) && startIndex < endIndex) {
            endIndex--;
        }

        String[] index = new String[2];
        index[0] = getLeftHandValue(startIndex);
        index[1] = getRightHandValue(startIndex, endIndex);

        logExit("getNextLine()");  //$NON-NLS-1$
        return index;
    }


    /**
     * Indicates if the line is empty, operates on mBOString
     *
     * @returns true if the line is empty, false otherwise
     * @param leftIndex
     * @param rightIndex
     * @param startIndex 
     */
    protected boolean emptyLine(int leftIndex,
                                int rightIndex,
                                int startIndex)
    {
        boolean result = true;

        while (Character.isWhitespace(boString.charAt(leftIndex)) && leftIndex < lineEndIndex) {
            leftIndex++;
        }

        while (Character.isWhitespace(boString.charAt(rightIndex)) && startIndex < rightIndex) {
            rightIndex--;
        }

        if (rightIndex > leftIndex) {
            result = false;
        }

        return result;
    }


    /**
     * local method: get string fragment of a line in front of
     *  equal sign
     *  return trimed string fragment
     *
     * @return String
     * @param startIndex 
     */
    protected String getLeftHandValue(int startIndex)
    {
        String result = null;

        int equalSignIndex = boString.indexOf('=', startIndex);
        result = boString.substring(startIndex, equalSignIndex);
        result = result.trim();
        return result;
    }


    /**
     * get string fragment of a line after the equal sign
     *
     *  @return trimed string fragment
     * @param endIndex
     * @param startIndex 
     */
    protected String getRightHandValue(int startIndex,
                                       int endIndex)
    {
        String result = null;

        int equalSignIndex = boString.indexOf('=', startIndex);
        result = boString.substring(equalSignIndex + 1, endIndex);
        result = result.trim();
        //trime right
        return result;
    }


    private void logEntry(String message)
    {
        if (LOGGER.isLoggable(Level.INFO)) {
            LOGGER.entering(CLASSNAME, message);
        }
    }


    private void logExit(String message)
    {
        if (LOGGER.isLoggable(Level.INFO)) {
            LOGGER.exiting(CLASSNAME, message);
        }
    }
}
